import java.awt.Color;
import java.awt.Font;
import java.awt.FontFormatException;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Shape;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import javax.swing.Timer;

class Pos
{
	public int col, row;
	
	public Pos( int c, int r ) {
		col = c;
		row = r;
	}
	
	public boolean equals( Object o ) {
		if( o instanceof Pos ) {
			Pos p = (Pos)o;
			return( p.col == col && p.row == row );
		} else {
			return false;
		}
	}
}

public class VT100
{
	protected Session parent;
	protected NVT nvt;
	
	// 畫面的寬與高
	protected int width, height;
	
	// 模擬螢幕的尺寸
	protected int maxrow, maxcol;
	protected int topmargin, buttommargin, leftmargin, rightmargin;
	
	// 字元的垂直與水平間距
	protected int fontverticalgap, fonthorizontalgap, fontdescentadj;
	
	// 預設為字內容
	protected byte default_text = 0;

	// 預設字型相關
	protected String defFontname = "Monospaced";
	protected int defFontsize = 18, defResolution = 72;
	
	// 各種字型參數
	protected Font font;
	protected int fontwidth, fontheight, fontdescent;
	protected int resolution, fontsize;
	
	// 預設編碼
	// protected String defEncoding = "BIG5-HKSCS";
	
	// 畫面
	protected BufferedImage bi;
	protected Graphics g;
	
	// 各種參數
	protected Resource resource;
	
	// 目前游標所在位址
	protected int ccol, crow;
	
	// 上次游標所在位置
	protected int lcol, lrow;
	
	// 文字是否到了畫面的最右端
	protected boolean linefull;
	
	// 儲存螢幕上的相關資訊
	protected byte[][] text;	// Server 傳來的原始資料
	protected char[][] ctext;	// 轉碼後的 utf-8 char
	protected int[][] mbc;		// multibyte character 的第幾個 byte
	
	// 是否被滑鼠選取
	protected boolean[][] selected;
	
	// 記錄螢幕上何處需要 repaint
	protected boolean[][] needRepaint;
	protected List repaintList;
	
	// 把從 nvt 來的資料暫存起來
	protected byte[] buf;
	protected int bufpos, buflen;
	
	// 預設顏色設定，前景、背景、游標色
	protected static byte defFg = 37;
	protected static byte defBg = 40;
	protected static byte defAttr = 0;
	
	// 預設游標顏色
	protected static Color cursorColor = Color.GREEN;
	
	// 各種屬性
	protected static final byte BOLD = 1;
	protected static final byte UNDERLINE = 8;
	protected static final byte BLINK = 16;
	protected static final byte REVERSE = 64;
	
	// ASCII
	protected static final byte BEL = 7;
	protected static final byte BS  = 8;
	protected static final byte TAB = 9;
	protected static final byte LF  = 10;
	protected static final byte VT  = 11;
	protected static final byte FF  = 12;
	protected static final byte CR  = 13;
	protected static final byte SO  = 14;
	protected static final byte SI  = 15;
	
	// 閃爍用
	protected int blinkcount;
	protected boolean textblink, cursorblink;
	
	// 調色盤
	protected static final Color[] normal_colors = {
			new Color( 0, 0, 0),
			new Color( 128, 0, 0 ),
			new Color( 0, 128, 0 ),
			new Color( 128, 128, 0),
			new Color( 0, 0, 128 ),
			new Color( 128, 0, 128 ),
			new Color( 0, 128, 128 ),
			new Color( 192, 192, 192 ),
			};
	
	protected static final Color[] highlight_colors = {
			new Color( 128, 128, 128),
			new Color( 255,   0,   0 ),
			new Color(   0, 255,   0 ),
			new Color( 255, 255,   0),
			new Color(   0,   0, 255 ),
			new Color( 255,   0, 255 ),
			new Color(   0, 255, 255 ),
			new Color( 255, 255, 255 ),
			};
	
	// 畫面上各位置的屬性及前景、背景色
	protected byte[][] attributes;
	protected byte[][] fgcolors, bgcolors;
	
	// 目前的屬性及前景、背景色
	protected byte cattribute;
	protected byte cfgcolor, cbgcolor;
	
	public static final int NUMERIC_KEYPAD = 1;
	public static final int APPLICATION_KEYPAD = 2;
	protected int keypadmode;
	
	// 判斷畫面上的網址
	protected int urlstate;
	protected Vector probablyurl;
	protected boolean[][] isurl;
	
	// 重繪用的 Timer
	protected Timer ti;

	protected Convertor conv;

	public void cleanScreen()
	{
		// FIXME: magic number
		g.setColor( Color.BLACK );
		g.fillRect( 0, 0, bi.getWidth(), bi.getHeight() );
	}
	
	public void repaintAll()
	{
		for(int i = 1; i <= maxrow; i++) {
			for(int j = 1; j <= maxcol; j++) {
				setRepaint( j, i );
			}
		}
	}
	
	public void updateImage( BufferedImage b )
	{
		bi = b;
		g = bi.createGraphics();
	}
	
	public void updateSize( int w, int h )
	{
		width = w;
		height = h;
	}
	
	public void updateFont()
	{
		FontMetrics fm;
		int fh, fw;
		
		fonthorizontalgap = resource.getIntValue( Resource.FONT_HORIZONTAL_GAP );
		fontverticalgap = resource.getIntValue( Resource.FONT_VERTICLAL_GAP );
		fontdescentadj = resource.getIntValue( Resource.FONT_DESCENT_ADJUST );

		fontsize = resource.getIntValue( Resource.FONT_SIZE );
		if( fontsize == 0 ) {
			// 按照螢幕的大小設定字型
			fw = width / maxcol - fonthorizontalgap;
			fh = height / maxrow - fontverticalgap;
			
			if( fh > 2 * fw ) {
				fh = 2 * fw;
			}
			fontsize = fh;
		}
		
		// 先使用預設字型，若需要再覆寫
		font = new Font( defFontname, Font.PLAIN, fontsize );
		if( resource.getBooleanValue(Resource.FONT_USE_CUSTOM_FONT) ) {
			try {
				File f = new File( resource.getStringValue(Resource.FONT_TTF_PATH));
				Font basefont = Font.createFont( Font.TRUETYPE_FONT, new FileInputStream(f) );
				font = basefont.deriveFont( Font.PLAIN, fontsize );
			} catch (FontFormatException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		g.setFont( font );
		
		fm = g.getFontMetrics();

		fontheight = fontsize;
		fontwidth = fontsize / 2;
		fontdescent = (int) (1.0 * fm.getDescent() / fm.getHeight() * fontsize);
		
		fontheight += fontverticalgap;
		fontwidth += fonthorizontalgap;
		fontdescent += fontdescentadj;
	}

	public void updateScreen()
	{
		cleanScreen();
		repaintAll();
	}
	
	protected char bytesToChar( byte[] b, int from, int limit )
	{
		// FIXME: 寫死成 big5
		return conv.big5BytesToUCS2( b[from], b[from + 1] );
	}

	/*
	// 不使用 buffer
	protected byte getNextByte()
	{
		int b;
		
		b = nvt.readByte();
		
		// 來自下層的連線中斷。
		if( b == -1 ) {
			loop = false;
			
			// TODO: 應該設為 optional 的
			// parent.showMessage("連線中斷！");
			
			parent.setIcon( ZTerm.ICON_CLOSED );
		}
		
		return (byte)b;
	}
	*/
	
	// 使用 buffer
	protected byte getNextByte()
	{
		// buffer 用光了，再跟下層拿。
		// 應該用 isBufferEmpty() 判斷的，但為了效率直接判斷。
		if( bufpos == buflen ) {
			try {
				buflen = nvt.readBytes( buf );
			} catch (IOException e) {
				// e.printStackTrace();
				// System.out.println("Connection closed.");
				parent.close( true );
			}
			bufpos = 0;
		}

		return buf[bufpos++];
	}
	
	// buffer 是否是空的
	public boolean isBufferEmpty()
	{
		return (bufpos == buflen);
	}
	
	// 設定某個位置需要重繪
	protected void setRepaint( int col, int row )
	{
		// XXX: 要檢查 row, col 的值是否合法，但我不知道這個狀況什麼時候會發生。
		if( row >= 1 && row <= maxrow && col >= 1 && col <= maxcol && !needRepaint[row - 1][col - 1] ) {
			needRepaint[row - 1][col - 1] = true;
			repaintList.add( new Pos( col, row ) );
		}
	}
	
	// 取得需要重繪的位置
	protected Pos getRepaint()
	{
		Pos p = (Pos)repaintList.get(0);
		needRepaint[p.row - 1][p.col - 1] = false;
		repaintList.remove(0);
		
		return p;
	}
	
	// 重繪佇列是否為空
	protected boolean isRepaintListEmpty()
	{
		return repaintList.isEmpty();
	}
	
	public void resetSelected()
	{
		for( int i = 1; i <= maxrow; i++) {
			for( int j = 1; j <= maxcol; j++) {
				selected[i - 1][j - 1] = false;
				setRepaint( j, i );
			}
		}
	}

	public void setSelected( int x1, int y1, int x2, int y2 )
	{
		int i, j;
		int r1, c1, r2, c2, tmp;
		
		c1 = x1 / fontwidth + 1;
		r1 = y1 / fontheight + 1;
		
		c2 = x2 / fontwidth + 1;
		r2 = y2 / fontheight + 1;

		if( r1 > r2 ) {
			tmp = r1;
			r1 = r2;
			r2 = tmp;
			
			tmp = c1;
			c1 = c2;
			c2 = tmp;
		} else if( r1 == r2 ) {
			if( c1 > c2 ) {
				tmp = c1;
				c1 = c2;
				c2 = tmp;
			}
		}
		
		for( i = 1; i <= maxrow; i++ ) {
			for( j = 1; j <= maxcol; j++ ) {
				
				if( i > r1 && i < r2 ) {
					selected[i - 1][j - 1] = true;
				} else if( i == r1 && i == r2 ) {
					selected[i - 1][j - 1] = (j >= c1 && j <= c2);
				} else if( i == r1 ) {
					selected[i - 1][j - 1] = (j >= c1);
				} else if( i == r2 ) {
					selected[i - 1][j - 1] = (j <= c2);
				} else {
					selected[i - 1][j - 1] = false;
				}
				
				// TODO: 應該改成 selected 狀態有更動的再重繪就好。
				setRepaint( j, i );		
			}
		}
	}
	
	public void selectConsequtive( int x, int y )
	{
		int c, r;
		int i, beginx, endx;
		
		c = x / fontwidth + 1;
		r = y / fontheight + 1;
		
		if( c <= 0 || c > maxcol || r <= 0 || r > maxrow ) {
			return;
		}
		
		// 往前找到第一個非空白的合法字元
		for( beginx = c; beginx > 0 && text[r - 1][beginx - 1] != default_text && text[r - 1][beginx - 1] != ' '; beginx-- );
		// 向後 ...
		for( endx = c; endx < maxcol && text[r - 1][endx - 1] != default_text && text[r - 1][endx - 1] != ' '; endx++ );
		
		resetSelected();
		for( i = beginx + 1; i < endx; i++ ) {
			selected[r - 1][i - 1] = true;
			setRepaint( i, r );
		}
	}
	
	public void selectEntireLine( int x, int y)
	{
		int i, r;
		
		r = y / fontheight + 1;
		
		resetSelected();
		for( i = 1; i < maxcol; i++ ) {
			selected[r - 1][i - 1] = true;
			setRepaint( i, r );
		}
	}
	
	public String getSelectedText()
	{	
		int i, j;
		boolean needNewLine = false, newLineAdded, currentLineHasSelected;
		StringBuffer sb = new StringBuffer();
		
		for( i = 1; i <= maxrow; i++ ) {
			currentLineHasSelected = newLineAdded = false;
			for( j = 1; j <= maxcol; j++ ) {
				if( selected[i - 1][j - 1] ) {
					
					if( needNewLine && (newLineAdded == false) ) {
						sb.append( "\n" );
						newLineAdded = true;
					}
					
					if( text[i - 1][j - 1] != default_text ) {
						if( isWideChar( text[i - 1][j - 1] ) && j < maxcol - 1 ) {
							sb.append( bytesToChar( text[i -1], j - 1, 2) );
							j++;
						} else {
							sb.append( (char) text[i - 1][j - 1]);
						}
					}
					currentLineHasSelected = true;
				} 
			}
			needNewLine = currentLineHasSelected;
		}
		
		return new String(sb);
	}
	
	public void pasteText( String str )
	{
		boolean autobreak;
		int count = 0, breakcount;
		char c;
		
		// 從系統取得的字串可能是 null
		if( str == null ) {
			return;
		}
		
		autobreak = resource.getBooleanValue( Resource.AUTO_LINE_BREAK );
		breakcount = resource.getIntValue( Resource.AUTO_LINE_BREAK_LENGTH );
		
		for( int i = 0; i < str.length(); i++ ) {
			c = str.charAt( i );
			
			if( c > 127 ) {
				parent.writeChar( c );
				count += 2;
			} else {
				if( c == 0x09 ) {
					// 遇到 tab 時加四
					// FIXME: magic number
					count += 4;
				} else if( c == 0x0a ) {
					// 遇到換行
					c = 0x0d;
					count = 0;
					
				} else {
					count++;
				}
				parent.writeChar( c );
			}
			
			// TODO: 斷行的演算法要再加強
			if( autobreak && count >= breakcount ) {
				parent.writeChar( (char)0x0d );
				count = 0;
			}
		}
	}

	protected void reset( int col, int row )
	{
		resetText( col, row );
		resetAttr( col, row );
		setRepaint( col, row );
	}
	
	protected void resetAttr( int col, int row )
	{
		fgcolors[row - 1][col - 1] = defFg;
		bgcolors[row - 1][col - 1] = defBg;
		attributes[row - 1][col - 1] = defAttr;
		
		isurl[row - 1][col - 1] = false;
	}
	
	protected void resetText( int col, int row )
	{
		text[row - 1][col - 1] = default_text;
		ctext[row - 1][col - 1] = (char)default_text;
		mbc[row - 1][col - 1] = 0;
	}

	protected void copy( int dstcol, int dstrow, int srccol, int srcrow )
	{
		text[dstrow - 1][dstcol - 1] = text[srcrow - 1][srccol - 1];
		ctext[dstrow - 1][dstcol - 1] = ctext[srcrow - 1][srccol - 1];
		mbc[dstrow - 1][dstcol - 1] = mbc[srcrow - 1][srccol - 1];

		fgcolors[dstrow - 1][dstcol - 1] = fgcolors[srcrow - 1][srccol - 1];
		bgcolors[dstrow - 1][dstcol - 1] = bgcolors[srcrow - 1][srccol - 1];
		attributes[dstrow - 1][dstcol - 1] = attributes[srcrow - 1][srccol - 1];
		
		isurl[dstrow - 1][dstcol - 1] = isurl[srcrow - 1][srccol - 1];

		setRepaint( dstcol, dstrow );
	}
	
	// 清除特定行
	protected void eraseline( int row, int mode ) {
		int i, begin, end;

		// System.out.println("erase line: " + row );
		
		switch( mode ) {
			case 0:
				begin = ccol;
				end = rightmargin;
				break;
			case 1:
				begin = leftmargin;
				end = ccol;
				break;
			case 2:
				begin = leftmargin;
				end = rightmargin;
				break;
			default:
				begin = leftmargin;
				end = rightmargin;
				break;
		}

		for( i = begin; i <= end; i++ ) {
			reset( i, row );
		}
	}
	
	// 清除螢幕
	protected void erasescreen( int mode ) {
		int i, begin, end;

		switch( mode ) {
			case 0:
				eraseline( crow, mode );
				begin = crow + 1;
				end = maxrow;
				break;
			case 1:
				eraseline( crow, mode );
				begin = 1;
				end = crow - 1;
				break;
			case 2:
				begin = 1;
				end = maxrow;
				break;
			default:
				begin = 1;
				end = maxrow;
				break;
		}

		for( i = begin; i <= end; i++) {
			eraseline( i, 2 );
		}
	}

	protected void insertline( int r, int n )
	{
		// System.out.println( "insert " + n + " line after " + r + " line");
		for(int i = buttommargin; i >= r; i--) {
			for(int j = leftmargin; j <= rightmargin; j++) {
				if( i >= r + n ) {
					copy( j, i, j, i - n);
				} else {
					reset( j, i );
				}
			}
		}
	}
	
	protected void reverseindex()
	{
		// System.out.println("reverse index at " + crow );
		if( crow == topmargin ) {
			insertline( crow, 1 );
		} else {
			crow--;
		}
	}
	
	protected void setmargin( int top, int buttom )
	{
		topmargin = top;
		buttommargin = buttom;
	}
	
	protected void scrollpage( int line )
	{
		Pos p;
		
		// System.out.println("scroll page " + line + " lines");
		for(int i = topmargin; i <= buttommargin; i++) {
			for(int j = leftmargin; j <= rightmargin; j++) {
				
				if( i <= buttommargin - line ) {
					copy( j, i, j, i + line);
				} else {
					reset( j, i );
				}

			}
		}
		
		for( int i = 0; i < probablyurl.size(); i++ ) {
			p = (Pos)probablyurl.elementAt( i );
			probablyurl.setElementAt( new Pos( p.col, p.row - line ), i );
		}
	}

	// TODO: 應該改用更好的寫法判斷是否為寬字元
	protected boolean isWideChar( byte a )
	{
		// XXX: 因為 byte 是 -128 ~ 127, 所以小於零表非 ASCII, 姑且就判斷是寬字元
		if( a < 0 )
			return true;
		return false;
	}

	protected void bell()
	{
		parent.bell();
	}
	
	public void paint( Graphics gg )
	{
		// TODO: 考慮 drawText 是否一定要擺在這邊，或是是否只在這裡呼叫？
		// 偶爾呼叫一次而不只是在顯示前才呼叫應該可以增進顯示速度。
		drawText();
		
		// 調整畫面大小，但速度極慢。
		// gg.drawImage( bi.getScaledInstance(1280, -1, Image.SCALE_FAST), 0, 0, null );
		gg.drawImage( bi, 0, 0, null );
	}

	protected void initResource()
	{
		// 讀入模擬終端機的大小，一般而言是 80 x 24
		maxcol = resource.getIntValue( Resource.TERMINAL_COLUMNS );
		maxrow = resource.getIntValue( Resource.TERMINAL_ROWS );
	}
	
	protected void initValue()
	{
		cfgcolor = defFg;
		cbgcolor = defBg;
		cattribute = defAttr;
		
		urlstate = 0;
		probablyurl = new Vector();
		
		blinkcount = 0;
		textblink = true;
		cursorblink = true;
		
		// 游標起始位址在螢幕左上角處
		ccol = 1;
		crow = 1;
		lcol = 1;
		lrow = 1;
		linefull = false;
		
		// 預設 margin 為整個螢幕
		topmargin = 1;
		buttommargin = maxrow;
		leftmargin = 1;
		rightmargin = maxcol;
		
		keypadmode = NUMERIC_KEYPAD;
	}
	
	protected void initArray()
	{
		text = new byte[maxrow][maxcol];
		ctext = new char[maxrow][maxcol];
		mbc = new int[maxrow][maxcol];
		
		selected = new boolean[maxrow][maxcol];
		needRepaint = new boolean[maxrow][maxcol];
		
		fgcolors = new byte[maxrow][maxcol];
		bgcolors = new byte[maxrow][maxcol];
		attributes = new byte[maxrow][maxcol];
		
		isurl = new boolean[maxrow][maxcol];
		
		// 初始化記載重繪位置用 LinkedList
		repaintList = Collections.synchronizedList(new LinkedList());
		
		for( int i = 0; i < maxrow; i++) {
			for( int j = 0; j < maxcol; j++) {
				needRepaint[i][j] = false;
				
				fgcolors[i][j] = defFg;
				bgcolors[i][j] = defBg;
				attributes[i][j] = defAttr;
				
				isurl[i][j] = false;
				
				reset( j + 1, i + 1);
			}
 		}
	}
	
	protected void initOthers()
	{
		// 啟動閃爍控制 thread
		ti = new Timer( 250, new RepaintTask() );
		ti.start();
	}
	
	public boolean coverURL( int x, int y )
	{
		int c = x / fontwidth, r = y / fontheight;
		
		if( r < maxrow && c < maxcol ) {
			return isurl[r][c];
		}
		
		return false;
	}
	
	public String getURL( int x, int y )
	{
		StringBuffer sb = new StringBuffer();
		int i;
		int c = x / fontwidth, r = y / fontheight;
		
		if( r >= maxrow && c >= maxcol ) {
			return new String();
		}

		for( i = c; i >= 0 && isurl[r][i]; i--);
		for( i++ ; i < maxcol && isurl[r][i]; i++ ) {
			sb.append( (char)text[r][i] );
		}

		return new String(sb);
	}
	
	protected void setColor( int c )
	{
		if ( c == 0 ) {
			cfgcolor = defFg;
			cbgcolor = defBg;
			cattribute = defAttr;
		} else if ( c == 1 ) {
			cattribute |= BOLD;
		} else if( c == 4 ) {
			cattribute |= UNDERLINE;
		} else if( c == 5 ) {
			cattribute |= BLINK;
		} else if ( c == 7 ) {
			cattribute |= REVERSE;
		} else if ( 30 <= c && c <= 37 ) {
			cfgcolor = (byte)c;
		} else if ( 40 <= c && c <= 47 ) {
			cbgcolor = (byte)c;
		}
	}

	public int getKeypadMode()
	{
		return keypadmode;
	}
	
	public Color getFgColor( int col, int row )
	{
		Color c;
		
		if( (attributes[row - 1][col - 1] & BOLD) != 0) {
			c = highlight_colors[ fgcolors[row - 1][col - 1] - 30 ];
		} else {
			c = normal_colors[ fgcolors[row - 1][col - 1] - 30 ];
		}
		
		if( (attributes[row - 1][col - 1] & REVERSE) != 0 ) {
			c = reverseColor( c );
		}
		
		return c;
	}
	
	public Color getBgColor( int col, int row )
	{
		Color c = normal_colors[ bgcolors[row - 1][col - 1] - 40 ];
		
		if( (attributes[row - 1][col - 1] & REVERSE) != 0 ) {
			c = reverseColor( c );
		}
		
		return c;
	}
	
	protected static Color reverseColor( Color c )
	{
		int r = c.getRed(), g = c.getGreen(), b = c.getBlue();
		
		r = 255 - r;
		g = 255 - g;
		b = 255 - b;

		return new Color( r, g, b);
	}
	
	protected void parse_csi()
	{
		int i, argc;
		int arg;
		int[] argv = new int[256];
		byte b;

		// System.out.print("CSI ");
		
		// 取得以 ';' 分開的參數
		arg = -1;
		argc = 0;
		while ( true ){

			b = getNextByte();
			
			if( '0' <= b && b <= '9' ) {
				if( arg == -1 ) {
					arg = 0;
				} else {
					arg *= 10;	
				}
				arg += b - '0';
			} else if( b == ';' ) {
				argv[argc] = arg;
				argc++;
				arg = 0;
			} else if( b == '?' || b == '!' || b == '"' || b == '\'') {
			} else {
				argv[argc] = arg;
				argc++;
				break;
			}
		}

		switch( b ) {
			case 'h': // TODO
				System.out.print( "Set mode(Not yet support): " );
				for( i = 0; i < argc; i++ ) {
					System.out.print( argv[i] + " " );
				}
				System.out.println();
				break;
			case 'm':
				for( i = 0; i < argc; i++ ) {
					if( argv[i] == -1 ) {
						argv[i] = 0;
					}
					setColor( argv[i] );
				}
				// System.out.println( "m" );
				break;
			case 'l': // TODO
				System.out.println( "Reset mode(Not yet support)" );
				break;
			case 'r': // TODO
				setmargin( argv[0], argv[1] );
				// System.out.println( "Set scroll margin: " + argv[0] + ", " + argv[1] );
				break;
			case 'A':
				if( argv[0] == -1 ) {
					argv[0] = 1;
				}
				crow = Math.max( crow - argv[0], topmargin );
				// System.out.println( argv[0] + " A" );
				break;
			case 'B':
				if( argv[0] == -1 ) {
					argv[0] = 1;
				}
				crow = Math.min( crow + argv[0], buttommargin );
				// System.out.println( argv[0] + " B" );
				break;
			case 'C':
				if( argv[0] == -1 ) {
					argv[0] = 1;
				}
				ccol = Math.min( ccol + argv[0], rightmargin );
				// System.out.println( argv[0] + " C" );
				break;
			case 'D':
				if( argv[0] == -1 ) {
					argv[0] = 1;
				}
				ccol = Math.max( ccol - argv[0], leftmargin );
				// System.out.println( argv[0] + " D" );
				break;
			case 'H':
				// ESC [ Pl ; Pc H
				// 移到 (pl, pc) 去，預設值是 (1, 1)
				// 座標是相對於整個畫面，不是 margin
				
				if( argv[0] < 1 ) {
					argv[0] = 1;
				}
				
				if( argv[1] < 1 ) {
					argv[1] = 1;
				}
				
				crow = Math.min( Math.max( argv[0], topmargin ), buttommargin );
				ccol = Math.min( Math.max( argv[1], leftmargin ), rightmargin );
				// System.out.println( argv[0] + " " + argv[1] + " H" );
				break;
			case 'J':
				if( argv[0] == -1 ) {
					argv[0] = 0;
				}
				erasescreen( argv[0] );
				// System.out.println( argv[0] + " J" );
				break;
			case 'K':
				if( argv[0] == -1 ) {
					argv[0] = 0;
				}
				eraseline( crow, argv[0] );
				// System.out.println( argv[0] + " K" );
				break;
			case 'L':
				if( argv[0] == -1 ) {
					argv[0] = 1;
				}
				insertline( crow, argv[0] );
				// System.out.println( argv[0] + " L" );
				break;
			default:
				System.out.println( "Unknown control sequence: " + b );
				break;
		}
	}
	
	protected void parse_text_parameter()
	{
		byte b;
		// FIXME: magic number
		byte[] text = new byte[80];
		int f, count;
		
		f = 0;
		b = getNextByte();
		while( b != ';' )
		{
			f *= 10;
			f += b - '0';
			b = getNextByte();
		}

		count = 0;
		b = getNextByte();
		while( b != BEL ) {
			text[count++] = b;
			b = getNextByte();
		}
		
		switch( f ) {
		case 0:
			parent.setIconName( new String( text, 0, count) );
			parent.setWindowTitle( new String( text, 0, count) );
			break;
		case 1:
			parent.setIconName( new String( text, 0, count) );
			break;
		case 2:
			parent.setWindowTitle( new String( text, 0, count) );
			break;
		default:
			System.out.println("Set text parameters(not fully support)");
			break;
		}
	}
	
	protected void parse_scs( byte a )
	{
		byte b;
		
		System.out.print( "SCS(Not yet support) " );
		
		b = getNextByte();
		
		System.out.println( a + " " + b );
	}
	
	protected void parse_esc()
	{
		byte b;

		// System.out.print("ESC ");

		b = getNextByte();
		
		switch( b ) {
			case '(': // 0x28
			case ')': // 0x29
				parse_scs( b );
				break;
			case '=': // 0x3d 
				keypadmode = APPLICATION_KEYPAD;
				// System.out.println( "Set application keypad mode" );
				break;
			case '>': // 0x3e
				keypadmode = NUMERIC_KEYPAD;
				// System.out.println( "Set numeric keypad mode" );
				break;
			case 'M': // 0x4d
				reverseindex();
				// System.out.println( "Reverse index" );
				break;
			case '[': // 0x5b
				parse_csi();
				break;
			case ']':
				parse_text_parameter();
				break;
			default:
				System.out.println( "Unknown control charecter: " + b );
				break;
		}
	}

	protected void parse()
	{
		byte b;
		boolean control;
		
		b = getNextByte();

		// 先把原來的游標位置存下來
		lcol = ccol;
		lrow = crow;
		
		control = true;
		
		switch( b ) {
			case 0:
			case 1:
			case 2:
			case 3:
			case 4:
			case 5:
			case 6:
				break;
			case 7: // Bell
				bell();
				
				// 如果 tab 不在最前則改變 icon
				if( !parent.isTabForeground() ) {
					parent.setIcon( ZTerm.ICON_ALERT );
				}
				
				break;
			case 8: // Backspace
				if( linefull ) {
					linefull = false;
				} else if( ccol > leftmargin ) {
					// 收到 Backspace 不需要清除文字，只要往前就好
					ccol--;
				} else if( ccol == leftmargin && crow > topmargin ) {
					ccol = rightmargin;
					crow--;
				}
				break;
			case 9: // TAB
				ccol = ((ccol - 1) / 8 + 1) * 8 + 1;
				break;
			case 10: // LF
				crow++;
				// 到 buttommargin 就該捲頁
				if( crow > buttommargin ) {
					scrollpage(1);
					crow--;
				}
				break;
			case 11:
				break;
			case 12:
				break;
			case 13: // CR
				ccol = leftmargin;
				break;
			case 14:
			case 15:
			case 16:
			case 17:
			case 18:
			case 19:
			case 20:
			case 21:
			case 22:
			case 23:
			case 24:
			case 25:
			case 26:
				break;
			case 27: // Esc
				parse_esc();
				break;
			case 28:
			case 29:
			case 30:
			case 31:
				break;
			default:
				control = false;
				break;
		}

		// 因控制碼會影響游標位置，所以只要游標不在行尾，linefull 就為 false.
		if( ccol != rightmargin ) {
			linefull = false;
		}
		
		// 上面 default case 的狀況，即一般字元，直接寫入。
		if( control == false ) {
			
			// 一般而言游標都在下一個字將會出現的地方，但若最後一個字在行尾（下一個字應該出現在行首）時，
			// 游標會在最後一個字上，也就是當最後一個字出現在行尾時並不會影響游標位置，
			// 游標會等到下一個字出現時才移到下行行首。
			if( linefull ) {
				linefull = false;
				ccol = leftmargin;
				crow++;
				if( crow > buttommargin ) {
					scrollpage(1);
					crow--;
				}
				// 游標會跳過 ( leftmargin, crow ) 所以手動 setRepaint
				setRepaint( leftmargin, crow );
			}
			
			// XXX: 
			// 如果是行首，則必定是 multobyte character 的第一個 byte
			// 如果不是行首，則看前一個 byte 是否非 ASCII 且為第一個 byte.
			// 此規則乃為 big5 量身訂做，若要支援其他編碼需再加寫其他程式碼。
		
			text[crow - 1][ccol - 1] = b;
			
			// isurl 不同於 color or attribute 不會繼承前一 byte
			isurl[crow - 1][ccol - 1] = false;
			
			// 每次讀入一個 byte 就檢查是否出現 url
			checkURL( (char) b );
			
			if( ccol == 1 ) {
				ctext[crow - 1][ccol - 1] = (char)b;
				mbc[crow - 1][ccol - 1] = 1;
			} else {
				if( text[crow - 1][ccol - 2] < 0 && mbc[crow - 1][ccol - 2] == 1 ) {
					ctext[crow - 1][ccol - 2] = bytesToChar( text[crow - 1], ccol - 2, 2 ) ;
					ctext[crow - 1][ccol - 1] = 0;
					mbc[crow - 1][ccol - 1] = 2;
				} else {
					ctext[crow - 1][ccol - 1] = (char)b;
					mbc[crow - 1][ccol - 1] = 1;
				}
			}
	
			fgcolors[crow - 1][ccol - 1] = cfgcolor;
			bgcolors[crow - 1][ccol - 1] = cbgcolor;
			attributes[crow - 1][ccol - 1] = cattribute;
			
			// 控制碼不會讓游標跑到 rightmargin 以後的地方，只有一般字元會，所以在這裡判斷 linefull 就可以了。
			ccol++;
			if( ccol > rightmargin ) {
				linefull = true;
				ccol--;
			}

		} else { // control = true
			// 很髒的作法，但是非一般字元也應中斷 url
			checkURL( (char) b );
		}
	
		// 游標經過的地方需要重繪
		setRepaint( lcol, lrow );
		if( lcol != ccol || lrow != crow ) {
			setRepaint( ccol, crow );
		}
	}

	protected void checkURL( char c )
	{
		String W = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ;/?:@&=+$,-_.!~*'()%#";
		Iterator iter;
		
		switch( urlstate ) {
		case 0:
			probablyurl.removeAllElements();
			if( c == 'h' ) {
				urlstate = 1;
				probablyurl.addElement( new Pos( ccol, crow ) );
			}
			break;
		case 1:
			if( c == 't' ) {
				urlstate = 2;
				probablyurl.addElement( new Pos( ccol, crow ) );
			} else {
				urlstate = 0;
			}
			break;
		case 2:
			if( c == 't' ) {
				urlstate = 3;
				probablyurl.addElement( new Pos( ccol, crow ) );
			} else {
				urlstate = 0;
			}
			break;
		case 3:
			if( c == 'p' ) {
				urlstate = 4;
				probablyurl.addElement( new Pos( ccol, crow ) );
			} else {
				urlstate = 0;
			}
			break;
		case 4:
			if( c == ':' ) {
				urlstate = 6;
				probablyurl.addElement( new Pos( ccol, crow ) );
			} else if( c == 's' ){
				urlstate = 5;
				probablyurl.addElement( new Pos( ccol, crow ) );
			} else {
				urlstate = 0;
			}
			break;
		case 5:
			if( c == ':' ) {
				urlstate = 6;
				probablyurl.addElement( new Pos( ccol, crow ) );
			} else {
				urlstate = 0;
			}
			break;
		case 6:
			if( c == '/' ) {
				urlstate = 7;
				probablyurl.addElement( new Pos( ccol, crow ) );
			} else {
				urlstate = 0;
			}
			break;
		case 7:
			if( c == '/' ) {
				urlstate = 8;
				probablyurl.addElement( new Pos( ccol, crow ) );
			} else {
				urlstate = 0;
			}
			break;
		case 8:
			if( W.indexOf(c) != -1 ) {
				urlstate = 9;
				probablyurl.addElement( new Pos( ccol, crow ) );
			} else {
				urlstate = 0;
			}
			break;
		case 9:
			if( W.indexOf(c) == -1 ) {
				iter = probablyurl.iterator();
				while( iter.hasNext() ) {
					Pos p = (Pos)iter.next();
					isurl[p.row - 1][p.col - 1] = true;
					setRepaint( p.col, p.row );
				}
				urlstate = 0;
			} else {
				probablyurl.addElement( new Pos( ccol, crow ) );
			}
			break;
		default:
			urlstate = 0;	
		}
	}

	public String makePasteText( Vector a, Vector b, Vector fg, Vector bg )
	{
		int i, j, c;
		
		byte cattr, tmpattr;
		byte cfg, cbg, tmpfg, tmpbg;
		byte mask;
		
		boolean needReset, needControl, isFirst;
		StringBuffer sb = new StringBuffer();
		
		// FIXME: magic number
		char[] buf = new char[32];
		
		buf[0] = 21;
		buf[1] = '[';
		buf[2] = '0';
		buf[3] = 'm';
		sb.append( buf, 0, 4 );
		
		cattr = 0;
		cfg = defFg;
		cbg = defBg;
		
		for(i = 0; i < a.size();i++) {
			
			tmpattr = ((Byte)a.elementAt(i)).byteValue();
			tmpfg = ((Byte)fg.elementAt(i)).byteValue();
			tmpbg = ((Byte)bg.elementAt(i)).byteValue();
			
			for( mask = 1, j = 0, needReset = needControl = false; j < 7; j++ ) {
				// 如果有屬性消失了，就得 reset
				if( ( (cattr & (mask << j)) != 0 ) && ( (tmpattr & (mask << j)) == 0 ) ) {
					needReset = true;
				}				
				
				// 如果有屬性改變，則需要加入控制碼
				if( (cattr & (mask << j)) != (tmpattr & (mask << j)) ) {
					needControl = true;
				}
				
				// 如果顏色變了，也需要加入控制碼
				if( cfg != tmpfg || cbg != tmpbg ) {
					needControl = true;
				}
			}
			
			// 顏色、屬性都沒變
			if( !needControl ) {
				sb.append( (char) ((Byte)b.elementAt(i)).byteValue() );
				continue;
			}
			
			// 如果需要控制碼才往下作
			
			buf[0] = 21;
			buf[1] = '[';

			if( needReset ) {
				buf[2] = '0';
				
				for(mask = 1, j = 0, c = 3; j < 7; j++) {					
					if( (tmpattr & (mask << j)) != 0 ) {
						buf[c++] = ';';
						buf[c++] = (char) ('1' + j);
					}
				}
				
				// 若是顏色跟預設的不同則需要設定
				if( tmpfg != defFg ) {
					buf[c++] = ';';
					buf[c++] = (char) ((tmpfg / 10) + '0');
					buf[c++] = (char) ((tmpfg % 10) + '0');
				}
				
				if( tmpbg != defBg ) {
					buf[c++] = ';';
					buf[c++] = (char) ((tmpbg / 10) + '0');
					buf[c++] = (char) ((tmpbg % 10) + '0');
				}
				
			} else {
				
				isFirst = true;
				for(mask = 1, j = 0, c = 2; j < 7; j++) {
					
					// 有新的屬性
					if( (tmpattr & (mask << j)) != 0 && (cattr & (mask << j)) == 0 ) {
						if( isFirst ) {
							isFirst = false;
						} else {
							buf[c++] = ';';
						}
						buf[c++] = (char) ('1' + j);
					}
				}
				
				if( cfg != tmpfg ) {
					if( isFirst ) {
						isFirst = false;
					} else {
						buf[c++] = ';';
					}
					buf[c++] = (char) ((tmpfg / 10) + '0');
					buf[c++] = (char) ((tmpfg % 10) + '0');
				}
				
				if( cbg != tmpbg ) {
					if( isFirst ) {
						isFirst = false;
					} else {
						buf[c++] = ';';
					}
					buf[c++] = (char) ((tmpbg / 10) + '0');
					buf[c++] = (char) ((tmpbg % 10) + '0');
				}
			}
			
			buf[c++] = 'm';
			buf[c++] = (char) ((Byte)b.elementAt(i)).byteValue();
			sb.append( buf, 0, c);
			
			cattr = tmpattr;
			cfg = tmpfg;
			cbg = tmpbg;
		}

		return new String(sb);
	}

	public String getSelectedColorText()
	{
		int i, j;
		Vector a = new Vector();	// attributes
		Vector b = new Vector();	// bytes
		Vector fg = new Vector();	// foreground color
		Vector bg = new Vector();	// background color
		boolean needNewLine;
		
		for( i = 1; i <= maxrow; i++ ) {
			needNewLine = false;
			for( j = 1; j <= maxcol; j++ ) {
				if( selected[i - 1][j - 1] ) {
					if( text[i - 1][j - 1] != default_text ) {
						a.addElement( new Byte(attributes[i - 1][j - 1]) );
						b.addElement( new Byte(text[i - 1][j - 1]) );
						fg.addElement( new Byte( fgcolors[i - 1][j - 1] ) );
						bg.addElement( new Byte( bgcolors[i - 1][j - 1] ) );
					}
					needNewLine = true;
				}
			}
			if( needNewLine ) {
				a.addElement( new Byte( (byte)0 ) );
				b.addElement( new Byte( (byte)0x0d ) );
				fg.addElement( new Byte( defFg ) );
				bg.addElement( new Byte( defBg ) );
			}
		}
		
		return makePasteText( a, b, fg, bg );
	}
	
	public void pasteColorText( String str )
	{
		byte[] tmp = new byte[str.length()];
		
		for( int i = 0; i < str.length(); i++) {
			tmp[i] = (byte)str.charAt(i);
		}
		
		parent.writeBytes( tmp, 0, tmp.length );
	}
	
	protected void drawText()
	{
		Pos p;
		int i, j;
		int w, h;
		
		while( !isRepaintListEmpty() ) {
			
			p = getRepaint();
			i = p.row - 1;
			j = p.col - 1;
			
			h = i * fontheight;
			w = j * fontwidth;
				
			// 設定為背景色
			if( selected[i][j] ) {
				g.setColor( reverseColor( getBgColor( p.col, p.row ) ) );
			} else {
				g.setColor( getBgColor( p.col, p.row ) );
			}
			
			// 如果是游標所在處就設定為游標色
			// 加上閃爍控制
			if( crow == p.row && ccol == p.col && cursorblink ) {
				g.setColor( cursorColor );
			}
			
			// 如果是游標所在處就畫出游標的顏色，否則是背景色
			g.fillRect( w, h, fontwidth, fontheight );

			// 畫出底線
			if( ((attributes[i][j] & UNDERLINE) != 0) || isurl[i][j] ) {
				if( isurl[i][j] ) {
					g.setColor( Color.ORANGE );
				} else {
					g.setColor( getFgColor( p.col, p.row ) );
				}
				g.drawLine( w, h + fontheight - 1, w + fontwidth - 1, h + fontheight - 1 );
			}

			// 空白不重繪，
			if( text[i][j] == default_text ) {
				continue;
			}

			// multibyte character 只在第一 byte 畫字
			/*
			if( mbc[i][j] > 1 ) {
				setRepaint( p.col - mbc[i][j] + 1, p.row );
				continue;
			}
			*/
			
			// 設為前景色
			if( selected[i][j]) {
				g.setColor( reverseColor( getFgColor( p.col, p.row ) ) );
			} else {
				g.setColor( getFgColor( p.col, p.row ) );
			}

			// 畫出文字
			// 加上閃爍控制
			// 加上雙色字
			if( ( (attributes[i][j] & BLINK) == 0) || textblink ) {
				/*
				g.drawChars( ctext[i], j, 1, w, h + fontheight - fontdescent);
				*/

				// 利用 clip 的功能，只畫出部份（半個）中文字。
				// XXX: 每個中文都會畫兩次，又有 clip 的 overhead, 效率應該會受到蠻大的影響
				Shape oldclip = g.getClip();
				g.clipRect( w, h, fontwidth, fontheight );
				g.drawChars( ctext[i], j - (mbc[i][j] - 1), 1, w - fontwidth * (mbc[i][j] - 1), h + fontheight - fontdescent);
				g.setClip( oldclip );
			}

		}
	}
	
	public void close()
	{
		// 停止重繪用的 timer
		ti.stop();
	}
	
	class RepaintTask implements ActionListener
	{
		public void actionPerformed(ActionEvent e)
		{
			blinkcount++;
			
			// FIXME: magic number
			// 游標閃爍
			if( blinkcount % 2 == 0 ) {
				cursorblink = !cursorblink;
				setRepaint( ccol, crow );
				parent.repaint();
			}
			
			// FIXME: magic number
			// 文字閃爍
			if( blinkcount % 3 == 0 ) {
				textblink = !textblink;
				for(int i = 1; i <= maxcol; i++) {
					for(int j = 1; j <= maxrow; j++) {
						if( (attributes[j - 1][i - 1] & BLINK) != 0 ) {
							setRepaint( i, j );
						}
					}
				}
				parent.repaint();
			}
		}
	}
	
	public VT100( int w, int h, Network n, Session p, BufferedImage b, Resource r, Convertor c )
	{
		nvt = new NVT( n );

		parent = p;
		resource = r;
		conv = c;
		
		// 使用 buffer
		buf = new byte[4096];
		bufpos = buflen = 0;

		// resource 會決定畫面大小，因此要在 initArray(), initValue() 之前完成
		initResource();
		
		initArray();
		initValue();
		initOthers();
		
		updateImage( b );
		updateSize( w, h );
		updateFont();
		updateScreen();
	}
}
